-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM
--
-- Restricted Materials of IBM
--
-- 5724-C51
--
-- (C) Copyright IBM Corp. 2004 All Rights Reserved.
--
-- US Government Users Restricted Rights - Use, duplication or
-- disclosure restricted by GSA ADP Schedule Contract with
-- IBM Corp.
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/asisp_checkashtransitions.sql
-- SID: 1.1
-- Date Modified: 04/09/24 08:52:36
--
-- Module Description
--   Defines the asisp_checkASHTransitions stored procedure
USE Object
GO

PRINT 'Creating procedure asisp_checkASHTransitions'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'asisp_checkASHTransitions' AND type = 'P')
  DROP PROCEDURE asisp_checkASHTransitions
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM
--
-- Restricted Materials of IBM
--
-- 5724-C51
--
-- (C) Copyright IBM Corp. 2004 All Rights Reserved.
--
-- US Government Users Restricted Rights - Use, duplication or
-- disclosure restricted by GSA ADP Schedule Contract with
-- IBM Corp.
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/asisp_checkashtransitions.sql
-- SID: 1.1
-- Date Modified: 04/09/24 08:52:36
-- 
-- Checks the transitions recorded in the AlertStateHistory table between the 
-- OldAlertStateID and AlertStateID columns. For any case where there is a sequence
-- that doesn't record the actual transition that occurred, the later record's 
-- OldAlertStateID column is fixed. So, if for a given object there are these
-- records:
--   ash_id   OldAlertStateID   AlertStateID
--        1                 3              2
--        2                 3              3
-- they will be changed as follows:
--   ash_id   OldAlertStateID   AlertStateID
--        1                 3              2
--        2                 2              3
--
-- Run with @fix=0 and @verbose=1 to produce a report listing the rows that need
-- correction without actually correcting them.
CREATE PROCEDURE asisp_checkASHTransitions
  @fix     TINYINT = 1, -- If 1 then an discrepancies are fixed, otherwise no action is taken
  @verbose TINYINT = 0  -- If >=1 then diagnostic messages are printed.
AS
  SET NOCOUNT ON
  
  DECLARE
    @AlertStateID         INT,
    @ash_id               INT,
    @cid                  ClassID,
    @err                  INT,
    @id                   ObjID,
    @now                  NVARCHAR(40),
    @num_checked          INT,
    @num_updates          INT,
    @OldAlertStateID      INT,
    @prev_AlertStateID    INT,
    @prev_ash_id          INT,
    @prev_OldAlertStateID INT,
    @rc                   INT,
    @upd_AlertStateID     INT,
    @upd_ash_id           INT,
    @upd_OldAlertStateID  INT


  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_checkASHTransitions: @fix=%d. Starting', 0, 1, @now, @fix) WITH NOWAIT
  END

  SET @num_checked = 0
  SET @num_updates = 0
  
  DECLARE history INSENSITIVE CURSOR FOR
    SELECT DISTINCT cid, id
    FROM AlertStateHistory
    ORDER BY cid, id
  OPEN history
  FETCH NEXT FROM history INTO @cid, @id
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    IF @num_checked > 0 AND @num_checked % 100 = 0 AND @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_checkASHTransitions: Checked %d objects so far', 0, 1, @now, @num_checked) WITH NOWAIT
    END
  
    SELECT @ash_id = MAX(ash_id) FROM AlertStateHistory WHERE cid = @cid AND id = @id
    SELECT @AlertStateID = AlertStateID, @OldAlertStateID = OldAlertStateID FROM AlertStateHistory WHERE ash_id = @ash_id
    IF @OldAlertStateID = @AlertStateID
    BEGIN
      -- Look back through AlertStateHistory for this object for a transition,
      -- and then make sure that transition is good.
      SET @prev_ash_id = NULL
      SELECT @prev_ash_id = MAX(ash_id) FROM AlertStateHistory WHERE ash_id < @ash_id AND cid = @cid AND id = @id
        AND AlertStateID <> OldAlertStateID
      IF @prev_ash_id IS NOT NULL
      BEGIN
        SELECT @prev_AlertStateID = AlertStateID, @prev_OldAlertStateID = OldAlertStateID
        FROM AlertStateHistory
        WHERE ash_id = @prev_ash_id
        IF @prev_AlertStateID <> @AlertStateID
        BEGIN
          -- We need a record to indicate the change from @prev_AlertStateID to @AlertStateID
          SELECT @upd_ash_id = MIN(ash_id)
          FROM AlertStateHistory
          WHERE ash_id > @prev_ash_id
            AND ash_id <= @ash_id
            AND cid = @cid
            AND id  = @id
            AND AlertStateID = @AlertStateID
          IF @upd_ash_id IS NOT NULL
          BEGIN
            IF @verbose >= 1
            BEGIN
              SELECT @upd_AlertStateID = AlertStateID, @upd_OldAlertStateID = OldAlertStateID
              FROM AlertStateHistory
              WHERE ash_id = @upd_ash_id
              SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
              RAISERROR('%s: asisp_checkASHTransitions: CURRENT:  cid=%4s id=%09d AlertStateID=%d OldAlertStateID=%d ash_id=%d', 0, 1, @now, @cid, @id, @AlertStateID, @OldAlertStateID, @ash_id) WITH NOWAIT
              RAISERROR('%s: asisp_checkASHTransitions: UPDATE:   cid=%4s id=%09d AlertStateID=%d OldAlertStateID=%d ash_id=%d', 0, 1, @now, @cid, @id, @upd_AlertStateID, @upd_OldAlertStateID, @upd_ash_id) WITH NOWAIT
              RAISERROR('%s: asisp_checkASHTransitions:                                              v-----UPDATE------^', 0, 1, @now) WITH NOWAIT
              RAISERROR('%s: asisp_checkASHTransitions: PREVIOUS: cid=%4s id=%09d AlertStateID=%d OldAlertStateID=%d ash_id=%d', 0, 1, @now, @cid, @id, @prev_AlertStateID, @prev_OldAlertStateID, @prev_ash_id) WITH NOWAIT
            END
            SET @num_updates = @num_updates + 1
            IF @fix = 1
            BEGIN
              UPDATE AlertStateHistory SET OldAlertStateID = @prev_AlertStateID WHERE ash_id = @upd_ash_id
              SELECT @err = @@ERROR, @rc = @@ROWCOUNT
              IF @err <> 0
              BEGIN
                SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
                RAISERROR('%s: asisp_checkASHTransitions: Error %d on update: UPDATE AlertStateHistory SET OldAlertStateID = %d WHERE ash_id = %d', 11, 1, @now, @err, @prev_AlertStateID, @upd_ash_id) WITH NOWAIT
                BREAK
              END
            END
          END
        END
      END
    END
    SET @num_checked = @num_checked + 1
    FETCH NEXT FROM history INTO @cid, @id
  END
  DEALLOCATE history
  
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_checkASHTransitions: Finished. #updates=%d, #checked=%d', 0, 1, @now, @num_updates, @num_checked) WITH NOWAIT
  END
EndOfProc:
GO
