--------------------------------------------------
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                     
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
--------------------------------------------------
-- File: archive/db/sql/tables/alertstatehistory.sql
-- SID: 1.3.4.15
-- Date Modified: 04/11/16 14:09:48
--
-- Module Description:
--   This script defines the AlertStateHistory table. This table holds a history of resource alert state changes.
-- 
-- Column Description:
--   cid             - Class id of the resource this AlertState change applies to.
--   id              - Instance id of the resource this AlertState change applies to.
--   ctime           - Time stamp when row was inserted.
--   AlertStateID    - The value of the AlertState that was appied to the resource.
--   evt_cid         - Event class id of the event that caused the AlertState change.
--   evt_id          - Event instance id of the event that caused this AlertState change.
--   evt_priority    - Event priority of the event that caused this AlertState change. 
--   evt_ctime       - Event create timestamp of the event that caused this AlertState change.
--   evt_detail      - Event detail of the event that caused this AlertState change.
--   OldAlertStateID - Previous AlertState value for resource this AlertState change applies to.
--   evt_name        - Name of the event that caused this AlertState change.
--   evt_alertstate  - The AlertState of the event that caused this AlertState change.
--   ash_id          - Identity value of this AlertStateHistory row.
-- sp_help AlertStateHistory
USE Object
GO

DECLARE @now NVARCHAR(40), @indid INT, @keycnt INT

IF NOT EXISTS (SELECT * FROM sysobjects WHERE name = 'AlertStateHistory' AND type = 'U')
BEGIN
  SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
  RAISERROR('%s: Creating table AlertStateHistory', 0, 1, @now) WITH NOWAIT
  CREATE TABLE AlertStateHistory
  (
    cid                           ClassID       NOT NULL,
    id                            ObjID         NOT NULL,
    ctime                         DATETIME      NOT NULL  CONSTRAINT ASH_ctime_DF DEFAULT (CURRENT_TIMESTAMP),
    AlertStateID                  TINYINT       NOT NULL,
    evt_cid                       ClassID       NOT NULL,
    evt_id                        ObjID         NOT NULL,
    evt_priority                  TINYINT       NOT NULL,
    evt_ctime                     DATETIME      NOT NULL,
    evt_detail                    NVARCHAR(255) NOT NULL,
    OldAlertStateID               TINYINT       NULL      CONSTRAINT ASH_OldAlertStateID_DF DEFAULT (NULL),
    evt_name                      ObjName       NULL      CONSTRAINT ASH_evt_name_DF DEFAULT (''),
    evt_alertstate                SMALLINT      NOT NULL,
    ash_id                        INTEGER       NOT NULL  IDENTITY(1,1)
  )
END

-- Update evt_alertstate to NOT NULL before subsequent primary key creation
IF EXISTS (SELECT * FROM syscolumns WHERE id = OBJECT_ID('AlertStateHistory') AND name = 'evt_alertstate' AND isnullable=1)
BEGIN
  SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
  RAISERROR('%s: Making evt_alertstate column non-nullable in AlertStateHistory table.', 0, 1, @now) WITH NOWAIT
  UPDATE AlertStateHistory
    SET evt_alertstate=0
    WHERE evt_alertstate IS NULL
  ALTER TABLE AlertStateHistory ALTER COLUMN evt_alertstate SMALLINT NOT NULL
END

-- See if the table structure is updated.
IF NOT EXISTS (SELECT * FROM syscolumns WHERE id = OBJECT_ID('AlertStateHistory') AND name = 'ash_id')
BEGIN
  SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
  RAISERROR('%s: Adding ash_id column to existing AlertStateHistory table.', 0, 1, @now) WITH NOWAIT
  ALTER TABLE AlertStateHistory ADD ash_id INT IDENTITY(1,1) NOT NULL
END

-- Check existing primary key. If it is not on the ash_id column, then drop it so it will be recreated below
IF EXISTS (SELECT * FROM sysobjects WHERE name = 'ASH_PK' AND type = 'K')
BEGIN
  SELECT @indid = indid, @keycnt = keycnt
  FROM sysindexes WHERE name = 'ASH_PK' AND id = OBJECT_ID('AlertStateHistory')
  
  IF @keycnt <> 1 OR INDEX_COL('AlertStateHistory', @indid, 1) <> 'ash_id'
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: Dropping old primary key on AlertStateHistory table.', 0, 1, @now) WITH NOWAIT
    ALTER TABLE AlertStateHistory DROP CONSTRAINT ASH_PK
  END
END

IF NOT EXISTS (SELECT * FROM sysobjects WHERE name = 'ASH_PK' AND type = 'K')
BEGIN
  SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
  RAISERROR('%s: Creating primary key for AlertStateHistory table.', 0, 1, @now) WITH NOWAIT
  ALTER TABLE AlertStateHistory ADD CONSTRAINT ASH_PK PRIMARY KEY CLUSTERED(ash_id)
END

-- Check existing unique index. If it is not correct, drop it so it will be recreated below
IF EXISTS (SELECT * FROM sysobjects WHERE name = 'ASH_by_cid_id_UQ' AND xtype = 'UQ')
BEGIN
  SELECT @indid = indid, @keycnt = keycnt
  FROM sysindexes WHERE name = 'ASH_by_cid_id_UQ' AND id = OBJECT_ID('AlertStateHistory')
  
  IF @keycnt <> 10 OR INDEX_COL('AlertStateHistory', @indid, 7) <> 'OldAlertStateID'
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: Dropping old unique index on AlertStateHistory table.', 0, 1, @now) WITH NOWAIT
    ALTER TABLE AlertStateHistory DROP CONSTRAINT ASH_by_cid_id_UQ
  END
END

IF NOT EXISTS (SELECT * FROM sysobjects WHERE name = 'ASH_by_cid_id_UQ' AND xtype = 'UQ')
BEGIN
  SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
  RAISERROR('%s: Creating unique constraint on AlertStateHistory columns (cid, id, ctime, evt_cid, evt_id, AlertStateID, OldAlertStateID, evt_alertstate, evt_priority)', 0, 1, @now) WITH NOWAIT
  ALTER TABLE AlertStateHistory ADD CONSTRAINT ASH_by_cid_id_UQ UNIQUE (cid, id, ctime, evt_cid, evt_id, AlertStateID, OldAlertStateID, evt_alertstate, evt_priority) WITH FILLFACTOR=50
END

SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
RAISERROR('%s: Finished defining/upgrading AlertStateHistory', 0, 1, @now) WITH NOWAIT
GO

-------------------------------------------------------------------------------
-- tI_AlertStateHistory

IF EXISTS (SELECT * FROM sysobjects WHERE name='tI_AlertStateHistory' AND type='TR')
BEGIN
  PRINT 'Dropping obsolete trigger tI_AlertStateHistory'
  DROP TRIGGER tI_AlertStateHistory
END
