-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
--
-- Module Description
--   Defines the link table and its associated triggers.
USE Object
GO

SET QUOTED_IDENTIFIER OFF 
GO
SET ANSI_NULLS ON 
GO

IF NOT EXISTS (SELECT * FROM sysobjects WHERE name = 'link' AND type = 'U')
BEGIN
  PRINT 'Creating table link'
  CREATE TABLE link
  (
    link_type LinkType  NOT NULL,
    src_cid   ClassID   NOT NULL,
    src_id    ObjID     NOT NULL,
    dst_cid   ClassID   NOT NULL,
    dst_id    ObjID     NOT NULL,
    ctime     DATETIME  NOT NULL CONSTRAINT link_ctime_DF   DEFAULT (CURRENT_TIMESTAMP),
    deleted   TINYINT   NOT NULL CONSTRAINT link_deleted_DF DEFAULT (0)
  ) ON [PRIMARY];
  ALTER TABLE link ADD CONSTRAINT link_PK PRIMARY KEY CLUSTERED
  (
    link_type,
    src_cid,
    src_id,
    dst_cid,
    dst_id
  ) ON [PRIMARY];
END
GO


---------------------------------------------------------------------------------------------------
-- tI_link
PRINT 'Creating trigger tI_link'

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tI_link' AND type = 'TR')
  DROP TRIGGER tI_link
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
CREATE TRIGGER tI_link ON link FOR INSERT
AS
  IF @@ROWCOUNT = 0
    RETURN

  -- Table for keeping track of inserted rows that have not been processed yet
  DECLARE @unprocessed TABLE
  (
    src_cid NVARCHAR(4) NOT NULL,
    src_id  INT         NOT NULL,
    dst_id  INT         NOT NULL
  )

  -- Table for keeping track of inserted rows that are being processed
  DECLARE @i TABLE
  (
    src_cid NVARCHAR(4) NOT NULL,
    src_id  INT         NOT NULL,
    dst_id  INT         NOT NULL
  )

  -- Table for computing inserts into lob_link
  DECLARE @tmp_loblink TABLE
  (
    src_id  INT         NOT NULL,
    dst_id  INT         NOT NULL,
    gen     INT         NOT NULL
  )
  
  DECLARE @err INT, @now NVARCHAR(40), @rc INT, @verbose TINYINT, @has_lobc TINYINT, @has_lobl TINYINT
  SET @verbose = 0

  IF EXISTS (SELECT 1 FROM inserted WHERE link_type = 'LOBC')
    SET @has_lobc = 1
  ELSE
    SET @has_lobc = 0

  IF EXISTS (SELECT 1 FROM inserted WHERE link_type = 'LOBL')
    SET @has_lobl = 1
  ELSE
    SET @has_lobl = 0

  IF @has_lobl = 1
  BEGIN
    -- Update the nchild column on the shortcut to match that of the destination LOB
    UPDATE lob SET
      nchild = dst.nchild
    FROM LineOfBusiness_C AS lob
    JOIN inserted AS i ON i.src_cid = lob.cid AND i.src_id = lob.id
    JOIN LineOfBusiness_C AS dst WITH (NOLOCK) ON dst.cid = i.dst_cid AND dst.id = i.dst_id
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      ROLLBACK TRANSACTION
      RETURN
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tI_link: Set nchild on %d LineOfBusiness_V record(s)', 0, 1, @now, @rc) WITH NOWAIT
    END
    
    -- Update LineOfBusiness_A. The following updates are made:
    -- 1. _AlertStateID for shortcuts is always set to that of the target LOB since by construction
    --    a shortcut always initially has the same AlertState as its target. The reason we make the
    --    update here is to provide quicker feedback to the user when creating a shortcut LOB.
    -- 2. _LinkedCID is always set based on the dst_cid of the new link.
    -- 3. The IsShortcut flag is set if the _LinkedCID='LOB'
    -- 4. The LinkedToLOBInCRL flag is set if _LinkedCID='LOB' and the LOB on the destination side
    --    of the link has its InCRL flag set.
    UPDATE lob SET
      _AlertStateID = CASE i.dst_cid WHEN 'LOB' THEN dst._AlertStateID ELSE 1 END,
      _flags =
        CASE i.dst_cid 
        WHEN 'LOB' THEN lob._flags 
          | 0x00020000 /* IsShortcut */
          | CASE ISNULL(dst._flags,0) & 0x00080000 WHEN 0 THEN 0 ELSE 0x00100000 /* LinkedToLOBInCRL */ END
        ELSE lob._flags
        END,
      _LinkedCID = i.dst_cid
    FROM LineOfBusiness_V AS lob
    JOIN inserted AS i ON i.src_cid = lob.cid AND i.src_id = lob.id
    LEFT OUTER JOIN LineOfBusiness_V AS dst WITH (NOLOCK) ON dst.cid = i.dst_cid AND dst.id = i.dst_id
    WHERE i.link_type = 'LOBL'
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      ROLLBACK TRANSACTION
      RETURN
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tI_link: Set LinkedCID and IsShortcut flag on %d LineOfBusiness_V record(s)', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  -- For LOBC links we insert lob_link records for LOBs on the dst side of the links
  IF @has_lobc = 1
  BEGIN
    INSERT INTO @unprocessed SELECT src_cid, src_id, dst_id FROM inserted WHERE link_type = 'LOBC'
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      ROLLBACK TRANSACTION
      RETURN
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tI_link: Number of LOBC links to process: %d', 0, 1, @now, @rc) WITH NOWAIT
    END
    
    
    -- Loop through each set of LOBC inserts for which there are no unprocessed parent rows  
    WHILE @rc > 0
    BEGIN
      DELETE FROM @i
      INSERT INTO @i
        SELECT u.src_cid, u.src_id, u.dst_id
        FROM @unprocessed AS u
          LEFT OUTER JOIN @unprocessed AS p ON p.dst_id = u.src_id AND p.src_cid <> 'LOBC'
        WHERE p.dst_id IS NULL
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @err <> 0
      BEGIN
        ROLLBACK TRANSACTION
        RETURN
      END
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tI_link: Number of LOBC links that can be processed due to no missing parent rows: %d', 0, 1, @now, @rc) WITH NOWAIT
        IF @verbose >= 2
          SELECT '@i' AS '@i', * FROM @i
      END
      
      DELETE FROM u FROM @unprocessed AS u JOIN @i AS i ON i.src_cid = u.src_cid AND i.src_id = u.src_id AND i.dst_id = u.dst_id
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @err <> 0
      BEGIN
        ROLLBACK TRANSACTION
        RETURN
      END
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tI_link: Removed %d row(s) from table with remaining links to handle', 0, 1, @now, @rc) WITH NOWAIT
      END
      
      -- Insert lob_link records with src_id=0 for any LOBs directly under the LOBC
      IF EXISTS (SELECT 1 FROM @i WHERE src_cid = 'LOBC')
      BEGIN
        INSERT INTO lob_link (src_id, dst_id, gen, phy_cid, phy_id)
          SELECT 0, i.dst_id, 1, NULL, 0
          FROM @i AS i
          WHERE i.src_cid = 'LOBC'
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of lob_link records for link_type=LOBC, src_cid=LOBC: %d', 0, 1, @now, @rc) WITH NOWAIT
        END
        
        -- Insert the src_id=0 records for any LOBs that are children of those inserted
        INSERT INTO lob_link (src_id, dst_id, gen, phy_cid, phy_id)
        SELECT DISTINCT 0, l.dst_id, l.gen+1, l.phy_cid, l.phy_id
        FROM @i AS i
          JOIN lob_link AS l ON i.dst_id = l.src_id
          LEFT OUTER JOIN lob_link AS e ON e.src_id = 0 AND e.dst_id = l.dst_id
        WHERE e.dst_id IS NULL
          AND i.src_cid = 'LOBC'
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of lob_link records for children of link_type=LOBC, src_cid=LOBC: %d', 0, 1, @now, @rc) WITH NOWAIT
        END

      END
      -- Handle LOB->LOB containment
      IF EXISTS (SELECT 1 FROM @i WHERE src_cid = 'LOB')
      BEGIN
        -- Insert a lob_link record for each new link with the src_id and dst_id values
        -- from the new link records.
        INSERT INTO lob_link (src_id, dst_id, gen, phy_cid, phy_id)
          SELECT i.src_id, i.dst_id, 1, NULL, 0
          FROM @i AS i
          WHERE i.src_cid = 'LOB'
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of lob_link records for link_type=LOBC, src_cid=LOB: %d', 0, 1, @now, @rc) WITH NOWAIT
        END

        -- Insert the "ancestor" links for the new records. We do this by looking up
        -- the existing lob_link records where the dst_id value corresponds to the src_id
        -- value of the new link.
        INSERT INTO lob_link
          SELECT ll.src_id, i.dst_id, ll.gen+1, NULL, 0
          FROM @i AS i
          JOIN lob_link AS ll ON ll.dst_id = i.src_id
          WHERE i.src_cid = 'LOB'
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of ancestor lob_link records for link_type=LOBC, src_cid=LOB: %d', 0, 1, @now, @rc) WITH NOWAIT
        END

        -- Insert the "descendent" links for the new records.
        INSERT INTO lob_link
          SELECT DISTINCT lp.src_id, lc.dst_id, lp.gen+lc.gen, lc.phy_cid, lc.phy_id
          FROM @i AS i
          JOIN lob_link AS lp ON lp.dst_id = i.dst_id
          JOIN lob_link AS lc ON lc.src_id = i.dst_id
          LEFT OUTER JOIN lob_link AS lprev ON lprev.src_id = lp.src_id AND lprev.dst_id = lc.dst_id
          WHERE lprev.src_id IS NULL
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of descendent lob_link records for link_type=LOBC, src_cid=LOB: %d', 0, 1, @now, @rc) WITH NOWAIT
        END

        -- A new child must also appear as if it were the child of any LOB linked to its parent
        INSERT INTO lob_link
          SELECT lobl.src_id, i.dst_id, 1 AS gen, NULL, 0
          FROM @i AS i
          JOIN link AS lobl ON lobl.dst_cid = i.src_cid AND lobl.dst_id = i.src_id
          WHERE lobl.link_type = 'LOBL'
            AND lobl.src_cid = 'LOB'
            AND lobl.deleted   = 0
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of direct shortcut lob_link records for link_type=LOBC, src_cid=LOB: %d', 0, 1, @now, @rc) WITH NOWAIT
        END

        -- Using temporary table to ensure no duplicates (src_id, dst_id) are inserted to lob_link 
        DELETE FROM @tmp_loblink
        INSERT INTO @tmp_loblink (src_id, dst_id, gen)
          SELECT DISTINCT lc.src_id, i.dst_id, lc.gen+1
          FROM @i AS i
          JOIN link AS lobl ON lobl.dst_cid = i.src_cid AND lobl.dst_id = i.src_id
          JOIN lob_link AS lc ON lc.dst_id = lobl.src_id
          LEFT OUTER JOIN lob_link AS lprev ON lprev.src_id = lc.src_id AND lprev.dst_id = i.dst_id
          WHERE lobl.link_type = 'LOBL'
            AND lobl.src_cid = 'LOB'
            AND lobl.deleted   = 0
            AND lc.src_id > 0
            AND lprev.src_id IS NULL
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of indirect shortcut lob_link records inserted into @tmp_loblink: %d', 0, 1, @now, @rc) WITH NOWAIT
        END

        -- Delete dups
        DELETE ll FROM @tmp_loblink AS ll WHERE gen > (
          SELECT MIN(gen) FROM @tmp_loblink 
            WHERE src_id = ll.src_id AND dst_id = ll.dst_id
        )
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of indirect shortcut lob_link duplicate records deleted from @tmp_loblink: %d', 0, 1, @now, @rc) WITH NOWAIT
        END

        INSERT INTO lob_link
          SELECT src_id, dst_id, gen, NULL, 0
          FROM @tmp_loblink
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of indirect shortcut lob_link records for link_type=LOBC, src_cid=LOB: %d', 0, 1, @now, @rc) WITH NOWAIT
        END

        -- Handle case where LOBL links are created before LOBC links
        UPDATE ll SET phy_cid = lobl.dst_cid, phy_id = lobl.dst_id
        FROM lob_link AS ll
        JOIN @i AS i ON i.dst_id = ll.dst_id
        JOIN link AS lobl ON lobl.src_id = i.dst_id
        WHERE lobl.link_type = 'LOBL'
          AND lobl.src_cid   = 'LOB'
          AND lobl.deleted   = 0
          AND ll.phy_cid IS NULL
        SELECT @err = @@ERROR, @rc = @@ROWCOUNT
        IF @err <> 0
        BEGIN
          ROLLBACK TRANSACTION
          RETURN
        END
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tI_link: Number of lob_link(phy_cid, phy_id) updates based on previously existing LOBL links: %d', 0, 1, @now, @rc) WITH NOWAIT
        END
      END
      
      SELECT @rc = COUNT(*) FROM @unprocessed
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tI_link: Number of LOBC links left to process: %d', 0, 1, @now, @rc) WITH NOWAIT
        IF @verbose >= 2
          SELECT '@unprocessed' AS '@unprocessed', * FROM @unprocessed
      END
    END -- Loop through each set of LOBC inserts for which there are no unprocessed parent rows  
  END -- For LOBC links we insert lob_link records for LOBs on the dst side of the links
  
  -- The remainder of the trigger is meaningful only for LOBL links
  IF NOT EXISTS (SELECT 1 FROM inserted WHERE link_type = 'LOBL')
  BEGIN
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tI_link: No LOBL links to process', 0, 1, @now) WITH NOWAIT
    END
    GOTO EndOfTrigger
  END

  -- Insert lob_link records based on indirect LOB->LOB relationships
  INSERT INTO lob_link (src_id, dst_id, gen, phy_cid, phy_id)
    SELECT lobl.src_id, i.src_id, ll.gen, ll.phy_cid, ll.phy_id
    FROM inserted AS i
      JOIN link AS lobc ON lobc.dst_id = i.src_id
      JOIN link AS lobl ON lobl.dst_id = lobc.src_id
      JOIN lob_link AS ll ON ll.dst_id = i.dst_id
      LEFT OUTER JOIN lob_link AS ex ON ex.dst_id = i.src_id AND ex.src_id = lobl.src_id
    WHERE ex.dst_id IS NULL
      AND i.link_type = 'LOBL'
      AND i.src_cid = 'LOB'
      AND i.dst_cid = 'LOB'
      AND lobc.link_type = 'LOBC'
      AND lobc.src_cid = 'LOB'
      AND lobc.dst_cid = 'LOB'
      AND lobc.deleted = 0
      AND lobl.link_type = 'LOBL'
      AND lobl.src_cid = 'LOB'
      AND lobl.dst_cid = 'LOB'
      AND lobl.deleted = 0
      AND ll.gen = 1
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    ROLLBACK TRANSACTION
    RETURN
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tI_link: Number of lob_link records inserted based on previously existing LOBL links: %d', 0, 1, @now, @rc) WITH NOWAIT
  END
  IF @rc > 0
  BEGIN
    INSERT INTO lob_link(src_id, dst_id, gen)
      SELECT DISTINCT lp.src_id, i.src_id, lp.gen+1
      FROM inserted AS i
      JOIN lob_link AS lc ON lc.dst_id = i.src_id
      JOIN lob_link AS lp ON lp.dst_id = lc.src_id
      WHERE i.link_type = 'LOBL'
        AND NOT EXISTS (SELECT * FROM lob_link AS lprev WHERE lprev.src_id = lp.src_id AND lprev.dst_id = i.src_id)
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      ROLLBACK TRANSACTION
      RETURN
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tI_link: Number of lob_link records inserted based on previously existing ancestor LOBL links: %d', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  -- Set the phy_cid and phy_id columns by following the logical links
  UPDATE lsrc SET
    phy_cid = i.dst_cid,
    phy_id  = i.dst_id
  FROM 
    lob_link AS lsrc
    JOIN inserted AS i ON i.src_id = lsrc.dst_id
  WHERE i.src_cid = 'LOB' AND i.link_type = 'LOBL'
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    ROLLBACK TRANSACTION
    RETURN
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tI_link: Number of lob_link(phy_cid, phy_id) updates based on new LOBL links: %d', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Handle LOB-LOB-LOBL links for the shortcut LOB
  INSERT INTO lob_link (src_id, dst_id, gen, phy_cid, phy_id)
  SELECT i.src_id AS src_id, l.dst_id AS dst_id, l.gen AS gen, l.phy_cid, l.phy_id
  FROM inserted AS i
    -- get the children of the target
    JOIN lob_link AS l ON l.src_id = i.dst_id
    -- get the generation of the shortcut
    JOIN lob_link AS g ON g.src_id = 0 AND g.dst_id = l.src_id
    -- prevent duplicates
    LEFT OUTER JOIN lob_link AS lprev ON lprev.src_id = i.src_id AND lprev.dst_id = l.dst_id
  WHERE i.src_cid = 'LOB' AND i.dst_cid = 'LOB' AND i.link_type = 'LOBL' AND lprev.src_id IS NULL
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    ROLLBACK TRANSACTION
    RETURN
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tI_link: Number of lob_link records inserted for shortcut-to-target children: %d', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Handle LOB-LOB-LOBL links for the parents of the shortcut
  INSERT INTO lob_link (src_id, dst_id, gen, phy_cid, phy_id)
  SELECT DISTINCT p.src_id, lobl.dst_id, p.gen + lobl.gen AS gen, lobl.phy_cid, lobl.phy_id
  FROM inserted AS i
    -- get the physical parent of the shortcut
    JOIN link AS l ON l.dst_cid = i.src_cid AND l.dst_id = i.src_id
    -- get the children of the shortcut (populated in the previous step)
    JOIN lob_link AS lobl ON lobl.src_id = i.src_id
    -- get the parents of the shortcut
    JOIN lob_link AS p ON p.dst_id = i.src_id AND p.src_id > 0
    -- prevent duplicates
    LEFT OUTER JOIN lob_link AS lprev ON lprev.src_id = p.src_id AND lprev.dst_id = lobl.dst_id
  WHERE i.src_cid = 'LOB' AND i.dst_cid = 'LOB' AND i.link_type = 'LOBL'
    AND l.link_type = 'LOBC'
    AND l.deleted = 0
    AND lprev.src_id IS NULL
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    ROLLBACK TRANSACTION
    RETURN
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tI_link: Number of lob_link records inserted for ancestors of shortcut-to-target children: %d', 0, 1, @now, @rc) WITH NOWAIT
  END

EndOfTrigger:
  IF @has_lobc = 1 OR @has_lobl = 1
  BEGIN
    INSERT INTO StagedLOB ( link_type, src_cid, src_id, dst_cid, dst_id, ctime )
    SELECT link_type, src_cid, src_id, dst_cid, dst_id, ctime
    FROM inserted
    WHERE link_type IN ( 'LOBC', 'LOBL')
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      ROLLBACK TRANSACTION
      RETURN
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tI_link: Inserted %d row(s) into StagedLOB', 0, 1, @now, @rc) WITH NOWAIT
    END
  END
  
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tI_link: Finished', 0, 1, @now) WITH NOWAIT
  END  
GO

---------------------------------------------------------------------------------------------------
-- tU_link 
PRINT 'Creating trigger tU_link'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link' AND type = 'TR')
  DROP TRIGGER tU_link
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
CREATE TRIGGER tU_link ON link FOR UPDATE
AS
  IF @@ROWCOUNT = 0
    RETURN

  DECLARE @err INT, @now NVARCHAR(40), @rc INT, @verbose TINYINT
  SET @verbose = 0

  IF @verbose >= 1
  BEGIN
    SELECT @rc = COUNT(*) FROM inserted
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tU_link: Starting. #rows=%d', 0, 1, @now, @rc) WITH NOWAIT
  END
  
  IF UPDATE(deleted) AND EXISTS (SELECT 1 FROM deleted WHERE link_type = 'LOBL')
  BEGIN
    -- If the link type is LOBL, remove any records from the MenuItemCache table for the LOB on the
    -- source side of the links and their ancestors. We need to do this because the MenuItems for
    -- the higher level LOB objects depend on their descendents and the physical resources those
    -- descendents are linked to.
    DELETE FROM c
    FROM lob_link AS ll
      JOIN inserted AS i ON i.src_id = ll.dst_id
      JOIN deleted AS d ON d.link_type = i.link_type AND d.src_cid = i.src_cid AND d.src_id = i.src_id
        AND d.dst_cid = i.dst_cid AND d.dst_id = i.dst_id
      JOIN MenuItemCache AS c ON c.obj_cid = 'LOB' AND (c.obj_id = ll.src_id OR c.obj_id = i.src_id)
      JOIN MenuItemVersion AS v ON v.verno = c.verno
    WHERE i.link_type = 'LOBL'
      AND i.deleted <> d.deleted -- doesn't matter whether delete or undelete, we still want to 
                                 -- get rid of any MenuItemCache/MenuItemCacheDetail rows
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link: Deleted %d row(s) from MenuItemCache', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- If an LOBL link is marked as deleted then all rows in lob_link where
    -- the object on the source side of the link is listed as a destination
    -- object need to have the phy_cid, phy_id columns nulled since the object
    -- is no longer linked
    UPDATE lob_link SET phy_cid = NULL, phy_id = 0
      FROM inserted AS i
      JOIN deleted AS d ON  d.link_type     = i.link_type
                        AND d.src_cid       = i.src_cid
                        AND d.src_id        = i.src_id
                        AND d.dst_cid       = i.dst_cid
                        AND d.dst_id        = i.dst_id
    WHERE d.link_type     = 'LOBL'
      AND d.deleted       = 0
      AND i.deleted       = 1
      AND lob_link.dst_id = i.src_id
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link: Update phy_cid to NULL on %d row(s) in lob_link', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- If an LOB-LOB-LOBL link is deleted, then the children of the target
    -- must be deleted as children from the shortcut.
    -- First, delete all of the children from the shortcut's parents.
    DELETE ld
    FROM inserted AS i
      JOIN deleted AS d ON d.link_type = i.link_type AND d.src_cid = i.src_cid AND d.src_id = i.src_id
        AND d.dst_cid = i.dst_cid AND d.dst_id = i.dst_id
      -- join to get the children of the shortcut
      JOIN lob_link AS l ON l.src_id = i.src_id
      -- join to get the parents of the shortcut
      JOIN lob_link AS lp ON lp.dst_id = i.src_id AND lp.src_id <> 0
      -- join to find all of the children that need to be deleted
      JOIN lob_link AS ld ON ld.dst_id = l.dst_id AND ld.src_id = lp.src_id
    WHERE i.src_cid = 'LOB' AND i.dst_cid = 'LOB' AND i.link_type = 'LOBL'
      AND i.deleted = 1 AND d.deleted = 0
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link: Deleted %d shortcut-to-target-children parent row(s) from lob_link', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- Now delete all of the children from the shortcut
    DELETE l
    FROM inserted AS i
      JOIN deleted AS d ON d.link_type = i.link_type AND d.src_cid = i.src_cid AND d.src_id = i.src_id
        AND d.dst_cid = i.dst_cid AND d.dst_id = i.dst_id
      -- join to get the children of the shortcut
      JOIN lob_link AS l ON l.src_id = i.src_id
    WHERE i.src_cid = 'LOB' AND i.dst_cid = 'LOB' AND i.link_type = 'LOBL'
      AND i.deleted = 1 AND d.deleted = 0
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link: Deleted %d shortcut-to-target-children row(s) from lob_link', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- When the last LOBL link to an object is deleted then the HasShortcuts flag bit
    -- needs to be turned off
    IF EXISTS
    (
      SELECT 1
      FROM inserted AS i
        JOIN deleted AS d ON d.link_type = i.link_type
                         AND d.src_cid   = i.src_cid
                         AND d.src_id    = i.src_id
                         AND d.dst_cid   = i.dst_cid
                         AND d.dst_id    = i.dst_id
      WHERE d.deleted = 0
        AND i.deleted = 1
        AND i.src_cid = 'LOB'
        AND i.dst_cid = 'LOB'
        AND i.link_type = 'LOBL'
    )
    BEGIN
      UPDATE lob SET _flags = _flags & ~262144 /* HasShortcuts=0x00040000 */
      FROM LineOfBusiness_V AS lob
      JOIN inserted AS i ON i.dst_cid = lob.cid AND i.dst_id = lob.id
      JOIN deleted  AS d ON d.link_type = i.link_type
                        AND d.src_cid   = i.src_cid
                        AND d.src_id    = i.src_id
                        AND d.dst_cid   = i.dst_cid
                        AND d.dst_id    = i.dst_id
      LEFT OUTER JOIN link AS l WITH (NOLOCK) ON l.link_type = i.link_type
                        AND l.src_cid   = i.src_cid
                        AND l.src_id    = i.src_id
                        AND l.dst_cid   = i.dst_cid
                        AND l.dst_id    = i.dst_id
                        AND l.link_type = 'LOBL'
                        AND l.deleted   = 0
      WHERE l.link_type IS NULL
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @err <> 0
      BEGIN
        ROLLBACK TRANSACTION
        RETURN
      END
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tU_link: Turned off HasShortcuts flag on %d LOB(s)', 0, 1, @now, @rc) WITH NOWAIT
      END
    END
    
    -- Handle undeletes of LOBL links to physical resources
    IF EXISTS
    (
      SELECT 1
      FROM inserted AS i
        JOIN deleted AS d ON d.link_type = i.link_type
                         AND d.src_cid   = i.src_cid
                         AND d.src_id    = i.src_id
                         AND d.dst_cid   = i.dst_cid
                         AND d.dst_id    = i.dst_id
      WHERE d.deleted = 1
        AND i.deleted = 0
    )
    BEGIN
      -- Reset the attributes on the LOB that track deletion
      UPDATE o SET
        _DeletedCID  =       '', 
        _DeletedID   =        0, 
        _DeletedName =       '', 
        _DeletedTime =     NULL,
        _LinkedCID   = i.dst_cid
      FROM LineOfBusiness_V AS o
      JOIN inserted         AS i ON i.src_id    = o.id
      JOIN deleted          AS d ON d.link_type = i.link_type
                                AND d.src_cid   = i.src_cid
                                AND d.src_id    = i.src_id
                                AND d.dst_cid   = i.dst_cid
                                AND d.dst_id    = i.dst_id
      WHERE d.link_type = 'LOBL'
        AND d.deleted   = 1
        AND i.deleted   = 0
        AND o.deleted   = 0
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tU_link: Updated %d row(s) in LineOfBusiness_V for undeletes of LOBL links', 0, 1, @now, @rc) WITH NOWAIT
      END

      -- Restore lob_link's phy_cid and phy_id that were cleared on LOBL deletion
      UPDATE ll
      SET
        phy_cid = i.dst_cid,
        phy_id  = i.dst_id
      FROM lob_link AS ll
      JOIN inserted AS i  ON i.src_id    = ll.dst_id 
      JOIN deleted  AS d  ON d.link_type = i.link_type
                        AND  d.src_cid   = i.src_cid
                        AND  d.src_id    = i.src_id
                        AND  d.dst_cid   = i.dst_cid
                        AND  d.dst_id    = i.dst_id
      WHERE d.link_type     = 'LOBL'
        AND d.deleted       = 1
        AND i.deleted       = 0
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tU_link: Restored phy_cid,phy_id on %d row(s) in lob_link for undeletes of LOBL links', 0, 1, @now, @rc) WITH NOWAIT
      END
    END -- Handle undeletes

    -- delete EVSS links to parent service when LOBL link is deleted
    -- find any parent service.  If none is found, nothing more to do  
    -- Note:  make the temp tables work tables
    CREATE TABLE #parents
    (
      org_parent_cid  ClassID NULL,
      org_parent_id   ObjID NULL,
      parent_cid      ClassID NULL,
      parent_id       ObjID NULL,
      service         ObjID NOT NULL DEFAULT 0,
      gen             INT NOT NULL DEFAULT 0,
      link_type       LinkType NULL DEFAULT NULL,
      done            INT NOT NULL DEFAULT 0 
    )

    CREATE TABLE #LobIds
    (  
       parent_cid  ClassID NULL,
       parent_id   ObjID NULL
    )

    CREATE TABLE #GetSubServiceUnderLobs
    (  parent_cid  ClassID NULL,
       parent_id   ObjID NULL
    )

    CREATE TABLE #SubServiceUnderLobs
    (
       org_parent_cid ClassID NULL, 
       org_parent_id ClassID NULL, 
       service_id ObjID NOT NULL DEFAULT 0
    )
 
    -- When LOBL link is deleted the LOBC link is also deleted, so find the deleted LOBC links
    -- to retrieve any parent service that may be higher in the hierarchy
    DECLARE @service ObjID
    INSERT INTO #LobIds (parent_cid, parent_id) 
      SELECT 'LOB', l.src_id
       FROM inserted AS i 
         JOIN deleted AS d ON d.link_type = i.link_type 
                          AND d.src_cid   = i.src_cid
                          AND d.src_id    = i.src_id
                          AND d.dst_cid   = i.dst_cid
                          AND d.dst_id    = i.dst_id
         JOIN link AS l ON l.dst_cid = i.src_cid
                       AND l.dst_id  = i.src_id 
       WHERE i.deleted = 1
         AND d.deleted = 0
         AND i.link_type = 'LOBL'
         AND l.deleted = 1
         AND l.link_type ='LOBC' 
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link: Found %d LOBC link(s) for delete LOBL links', 0, 1, @now, @rc) WITH NOWAIT
    END
      
    IF EXISTS (SELECT 1 FROM #LobIds )
    BEGIN
      -- call find_parentservice with #LobIds as the input table get a list of the parent services 
      INSERT INTO #parents
        EXEC find_parentservice NULL, @parent = @service OUTPUT, @returnall = 1, @verbose = @verbose
    END

    --  If there are any parents services then links to parent services must be deleted
    IF EXISTS (SELECT 1 FROM #parents)
    BEGIN 
      --if service found in Linked LOB. Delete link from parent to service.
      UPDATE e
      SET deleted = 1
      FROM EXES_EXES_EVSS AS e
        JOIN #parents AS p ON p.service = e.src_id
        JOIN inserted AS i ON i.deleted = 1 
        JOIN deleted AS d ON d.link_type = i.link_type 
                         AND d.src_cid   = i.src_cid
                         AND d.src_id    = i.src_id
                         AND d.dst_cid   = i.dst_cid
                         AND d.dst_id    = i.dst_id
        JOIN LineOfBusiness_V AS v ON v.id = i.dst_id
        JOIN link AS l ON l.dst_id = v.id
        JOIN link AS l2 ON l2.dst_id = i.src_id
      WHERE e.src_id = p.service
        AND e.dst_id = l.src_id  
        AND i.deleted = 1
        AND d.deleted = 0
        AND i.link_type = 'LOBL'
        AND l.link_type = 'EVSL'  
        AND l.deleted = 0
        AND v.deleted = 0
        AND l2.link_type = 'LOBC'
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tU_link: Marked %d row(s) in EXES_EXES_EVSS for deletion', 0, 1, @now, @rc) WITH NOWAIT
      END

      --  Get any services that are children of the Linked LOB. 
      INSERT INTO #GetSubServiceUnderLobs
        SELECT 'LOB', i.dst_id
        FROM inserted AS i
          JOIN deleted AS d ON d.link_type = i.link_type 
                           AND d.src_cid   = i.src_cid
                           AND d.src_id    = i.src_id
                           AND d.dst_cid   = i.dst_cid
                           AND d.dst_id    = i.dst_id
          JOIN LineOfBusiness_V AS v ON  v.id = i.dst_id
          LEFT OUTER JOIN link AS evsl WITH (NOLOCK) ON evsl.link_type = 'EVSL' AND evsl.dst_id = v.id
        WHERE i.deleted = 1 
          AND d.deleted = 0
          AND v.deleted = 0
          AND evsl.dst_id IS NULL
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tU_link: Found %d subservice(s) under deleted LOBs', 0, 1, @now, @rc) WITH NOWAIT
      END

      -- Look for subservices at lower level. EVS_GetSubServiceUnderLobs is the input table when calling asisp_getsubservicesunderlob
      -- Delete links from parent service.
      IF EXISTS (SELECT 1 FROM #GetSubServiceUnderLobs)
      BEGIN 
        INSERT INTO #SubServiceUnderLobs EXEC asisp_getsubservicesunderlob NULL,1
        IF @verbose >= 1
        BEGIN
          SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
          RAISERROR('%s: tU_link: Resolved subservice chain for deleted LOBs', 0, 1, @now) WITH NOWAIT
        END

        --select * from  EVS_SubServiceUnderLobs
        IF EXISTS (SELECT 1 FROM #SubServiceUnderLobs) 
        BEGIN
          UPDATE e SET deleted = 1 
          FROM EXES_EXES_EVSS AS e
            JOIN #parents AS p ON p.service = e.src_id 
            JOIN #SubServiceUnderLobs AS s ON s.service_id = e.dst_id
            JOIN inserted AS i ON i.dst_id = s.org_parent_id
            JOIN deleted AS d ON d.src_id = i.src_id AND d.dst_id = i.dst_id AND d.link_type = i.link_type    
          WHERE d.deleted = 0
            AND i.deleted = 1          
          SELECT @err = @@ERROR, @rc = @@ROWCOUNT
          IF @verbose >= 1
          BEGIN
            SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
            RAISERROR('%s: tU_link: Marked %d row(s) in EXES_EXES_EVSS for deletion', 0, 1, @now, @rc) WITH NOWAIT
          END
        END 
      END
    END 

  END -- IF UPDATE(deleted) AND EXISTS (SELECT 1 FROM deleted WHERE link_type = 'LOBL')

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tU_link: Finished', 0, 1, @now) WITH NOWAIT
  END
-- END OF tU_link
GO

---------------------------------------------------------------------------------------------------
-- tU_link_unregister 
PRINT 'Creating trigger tU_link_unregister'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_unregister' AND type = 'TR')
  DROP TRIGGER tU_link_unregister
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- Update trigger on link table that calls the asisp_unregisterobjects
-- procedure for each LOBL link to a FILE or TRAN that is being deleted.
-- The asisp_unregisterobjects will unregister the FILE or TRAN if there
-- are no other LOBL links to it.
CREATE TRIGGER tU_link_unregister ON link FOR UPDATE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  DECLARE @rc INT
  IF UPDATE(deleted) AND EXISTS (SELECT 1 FROM deleted WHERE link_type = 'LOBL' AND dst_cid IN ( 'FILE', 'TRAN' ) )
  BEGIN
    -- Perform transaction and file unregistration
    SELECT @rc = COUNT(*)
    FROM inserted AS i
    JOIN deleted  AS d ON d.link_type = i.link_type
                      AND d.src_cid   = i.src_cid
                      AND d.src_id    = i.src_id
                      AND d.dst_cid   = i.dst_cid
                      AND d.dst_id    = i.dst_id
    WHERE
      i.link_type ='LOBL'
      AND i.dst_cid IN ( 'FILE', 'TRAN' )
      AND i.deleted = 1
      AND d.deleted = 0

    -- If only one row being marked for deletion run asisp_unregisterobjects for the specific
    -- object on the dst side of the link.
    IF @rc = 1  
    BEGIN
      DECLARE @dst_cid ClassID, @dst_id ObjID
      SELECT @dst_cid = d.dst_cid, @dst_id = d.dst_id
      FROM inserted AS i
      JOIN deleted  AS d ON d.link_type = i.link_type
                        AND d.src_cid   = i.src_cid
                        AND d.src_id    = i.src_id
                        AND d.dst_cid   = i.dst_cid
                        AND d.dst_id    = i.dst_id
      WHERE
        i.link_type ='LOBL'
        AND i.dst_cid IN ( 'FILE', 'TRAN' )
        AND i.deleted = 1
        AND d.deleted = 0
      EXEC asisp_unregisterobjects @dst_cid, @dst_id
    END
    ELSE IF @rc > 1
    BEGIN
      -- More than one row affected, so run the unregister procedures generally (without specifying the instance id)
      IF EXISTS
      (
        SELECT 1
        FROM inserted AS i
        JOIN deleted  AS d ON d.link_type = i.link_type
                          AND d.src_cid   = i.src_cid
                          AND d.src_id    = i.src_id
                          AND d.dst_cid   = i.dst_cid
                          AND d.dst_id    = i.dst_id
        WHERE
          i.link_type ='LOBL'
          AND i.dst_cid = 'TRAN'
          AND i.deleted = 1
          AND d.deleted = 0
      )
        EXEC asisp_unregisterobjects 'TRAN'

      IF EXISTS
      (
        SELECT 1
        FROM inserted AS i
        JOIN deleted  AS d ON d.link_type = i.link_type
                          AND d.src_cid   = i.src_cid
                          AND d.src_id    = i.src_id
                          AND d.dst_cid   = i.dst_cid
                          AND d.dst_id    = i.dst_id
        WHERE
          i.link_type ='LOBL'
          AND i.dst_cid = 'FILE'
          AND i.deleted = 1
          AND d.deleted = 0
      )
        EXEC asisp_unregisterobjects 'FILE'
    END -- Perform transaction and file unregistration
  END -- IF UPDATE(deleted) AND EXISTS (SELECT 1 FROM deleted WHERE link_type = 'LOBL' AND dst_cid IN ( 'FILE', 'TRAN' ) )
-- END OF tU_link_unregister
GO

---------------------------------------------------------------------------------------------------
-- tI_link_IMS_check 
IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tI_link_IMS_check' AND type = 'TR')
BEGIN
  PRINT 'Removing obsolete trigger tI_link_IMS_check'
  DROP TRIGGER tI_link_IMS_check
END
GO

---------------------------------------------------------------------------------------------------
-- tU_link_IMS_check
IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_IMS_check' AND type = 'TR')
BEGIN
  PRINT 'Removing obsolete trigger tU_link_IMS_check'
  DROP TRIGGER tU_link_IMS_check
END
GO


---------------------------------------------------------------------------------------------------
-- tD_link_cache
PRINT 'Creating trigger tD_link_cache'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tD_link_cache' AND type = 'TR')
  DROP TRIGGER tD_link_cache
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- Delete trigger on the link table for keeping the ObjPathCacheLinks table in sync.
-- If any record is being deleted from the link table with the deleted column equal to zero,
-- then a corresponding record must exist in ObjPathCacheLinks to indicate this deletion,
-- otherwise the ObjPathCache table may end up with bogus records based on the deleted link.
-- Note that under normal circumstances this would not happen because links are first marked
-- as deleted, and in that case the tU_link_cache trigger takes care of things.
CREATE TRIGGER tD_link_cache ON link FOR DELETE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  IF EXISTS
  (
    SELECT 1
    FROM deleted AS d
      JOIN isa_chain AS ic WITH (NOLOCK) ON ic.cid = d.dst_cid
      JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = d.link_type
    WHERE ic.base_cid = 'MNO2'
      AND (lt.flag_containment = 1 OR lt.flag_parent_child = 1 OR lt.flag_propagation = 1 OR (d.link_type = 'LOBL' AND d.src_cid = 'LOB' AND d.dst_cid = 'LOB'))
      AND lt.flag_aggregate = 0
      AND d.deleted = 0
  )
  BEGIN

    -- first check if the values are already in the ObjPathCacheLinks table
    UPDATE o
    SET deleted = 1
    FROM ObjPathCacheLinks AS o
    JOIN deleted AS d ON o.dst_cid   = d.dst_cid
                     AND o.dst_id    = d.dst_id
                     AND o.src_cid   = d.src_cid
                     AND o.src_id    = d.src_id
                     AND o.link_type = d.link_type
    WHERE d.deleted = 0

    -- add any entries that are not in the cache
    INSERT INTO ObjPathCacheLinks (link_type, src_cid, src_id, dst_cid, dst_id, deleted)
    SELECT RTRIM(d.link_type), RTRIM(d.src_cid), d.src_id, RTRIM(d.dst_cid), d.dst_id, 1
    FROM ObjPathCacheLinks AS o
      RIGHT OUTER JOIN deleted AS d ON d.dst_cid   = o.dst_cid
                                   AND d.dst_id    = o.dst_id
                                   AND d.src_cid   = o.src_cid
                                   AND d.src_id    = o.src_id
                                   AND d.link_type = o.link_type
      JOIN ObjPathCacheIDs AS opc ON opc.cid = d.dst_cid
      JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = d.link_type
      JOIN isa_chain AS ic WITH (NOLOCK) ON ic.cid = d.dst_cid
    WHERE o.dst_cid IS NULL
      AND ic.base_cid = 'MNO2'
      AND ((d.dst_id <= opc.id AND d.link_type <> 'LOBL') OR (d.link_type = 'LOBL' AND d.src_id <= opc.id))
      AND d.deleted = 0

  END
-- END OF tD_link_cache
GO

---------------------------------------------------------------------------------------------------
-- tI_link_cache
PRINT 'Creating trigger tI_link_cache'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tI_link_cache' AND type = 'TR')
  DROP TRIGGER tI_link_cache
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
CREATE TRIGGER tI_link_cache ON link FOR INSERT WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN
--RAISERROR('tI_link_cache', 0, 1) WITH NOWAIT
  IF EXISTS
  (
    SELECT 1
    FROM inserted AS i
      JOIN isa_chain AS ic WITH (NOLOCK) ON ic.cid = i.dst_cid
      JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = i.link_type
    WHERE ic.base_cid = 'MNO2'
      AND (lt.flag_containment = 1 OR lt.flag_parent_child = 1 OR lt.flag_propagation = 1 OR (i.link_type = 'LOBL' AND i.src_cid = 'LOB' AND i.dst_cid = 'LOB'))
      AND lt.flag_aggregate = 0
      AND i.deleted = 0
  )
  BEGIN

    -- if any of the entries already exist, just set the deleted flag to 1
    UPDATE o
    SET deleted = 0
    FROM ObjPathCacheLinks AS o
      JOIN inserted AS i ON i.dst_cid   = o.dst_cid
                        AND i.dst_id    = o.dst_id
                        AND i.src_cid   = o.src_cid
                        AND i.src_id    = o.src_id
                        AND i.link_type = o.link_type
    WHERE i.deleted = 0
      AND o.deleted = 1

    -- add the new entries
    INSERT INTO ObjPathCacheLinks (link_type, src_cid, src_id, dst_cid, dst_id, deleted)
    SELECT RTRIM(i.link_type), RTRIM(i.src_cid), i.src_id, RTRIM(i.dst_cid), i.dst_id, i.deleted
    FROM ObjPathCacheLinks AS o
      RIGHT OUTER JOIN inserted AS i ON i.dst_cid   = o.dst_cid
                                    AND i.dst_id    = o.dst_id
                                    AND i.src_cid   = o.src_cid
                                    AND i.src_id    = o.src_id
                                    AND i.link_type = o.link_type
      JOIN ObjPathCacheIDs AS opc ON opc.cid = i.dst_cid
      JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = i.link_type
      JOIN isa_chain AS ic WITH (NOLOCK) ON ic.cid = i.dst_cid
    WHERE i.deleted = 0
      AND ((i.dst_id <= opc.id AND i.link_type <> 'LOBL') OR (i.link_type = 'LOBL' AND i.src_id <= opc.id))
      AND ic.base_cid = 'MNO2'
      AND (lt.flag_containment = 1 OR lt.flag_parent_child = 1 OR lt.flag_propagation = 1 OR (i.link_type = 'LOBL' AND i.src_cid = 'LOB' AND i.dst_cid = 'LOB'))
      AND lt.flag_aggregate = 0
      AND o.dst_cid IS NULL

  END
-- END OF tI_link_cache
GO

---------------------------------------------------------------------------------------------------
-- tU_link_OS_CICS_RUNC
PRINT 'Creating trigger tU_link_OS_CICS_RUNC'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_OS_CICS_RUNC' AND type = 'TR')
  DROP TRIGGER tU_link_OS_CICS_RUNC
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- Update trigger on link table for propagating deleted flag changes to the OS_CICS_RUNC table
CREATE TRIGGER tU_link_OS_CICS_RUNC ON link FOR UPDATE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  IF UPDATE(deleted) AND EXISTS
  (
    SELECT 1
    FROM deleted AS d
    JOIN inserted AS i ON i.link_type = d.link_type
                      AND i.src_cid   = d.src_cid
                      AND i.src_id    = d.src_id
                      AND i.dst_cid   = d.dst_cid
                      AND i.dst_id    = d.dst_id
    WHERE d.src_cid   = 'OS'
      AND d.dst_cid   = 'CICS'
      AND d.link_type = 'RUNC'
      AND i.deleted   <> d.deleted
  )
  BEGIN
    UPDATE il SET deleted = i.deleted
    FROM OS_CICS_RUNC AS il
      JOIN deleted AS d ON d.src_cid = 'OS' AND d.dst_cid = 'CICS' AND d.link_type = 'RUNC' AND d.src_id = il.src_id AND d.dst_id = il.dst_id
      JOIN inserted AS i ON i.link_type = d.link_type AND i.src_cid = d.src_cid AND i.src_id = d.src_id AND i.dst_cid = d.dst_cid AND i.dst_id = d.dst_id
    WHERE
      i.deleted <> d.deleted
      AND il.deleted <> i.deleted
  END
-- END OF tU_link_OS_CICS_RUNC
GO

---------------------------------------------------------------------------------------------------
-- tU_link_CICF_CICS_MAGC
PRINT 'Creating trigger tU_link_CICF_CICS_MAGC'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_CICF_CICS_MAGC' AND type = 'TR')
  DROP TRIGGER tU_link_CICF_CICS_MAGC
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- Update trigger on link table for propagating deleted flag changes to the CICF_CICS_MAGC table
CREATE TRIGGER tU_link_CICF_CICS_MAGC ON link FOR UPDATE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  IF UPDATE(deleted) AND EXISTS
  (
    SELECT 1
    FROM deleted AS d
      JOIN inserted AS i ON i.link_type = d.link_type
                        AND i.src_cid   = d.src_cid
                        AND i.src_id    = d.src_id
                        AND i.dst_cid   = d.dst_cid
                        AND i.dst_id    = d.dst_id
    WHERE d.src_cid   = 'CICF'
      AND d.dst_cid   = 'CICS'
      AND d.link_type = 'MAGC'
      AND i.deleted   <> d.deleted
  )
  BEGIN
    UPDATE il SET deleted = i.deleted
    FROM CICF_CICS_MAGC AS il
      JOIN deleted AS d ON d.src_cid = 'CICF' AND d.dst_cid = 'CICS' AND d.link_type = 'MAGC' AND d.src_id = il.src_id AND d.dst_id = il.dst_id
      JOIN inserted AS i ON i.link_type = d.link_type AND i.src_cid = d.src_cid AND i.src_id = d.src_id AND i.dst_cid = d.dst_cid AND i.dst_id = d.dst_id
    WHERE
      i.deleted <> d.deleted
      AND il.deleted <> i.deleted
  END
-- END OF tU_link_CICF_CICS_MAGC
GO

---------------------------------------------------------------------------------------------------
-- tU_link_cache
PRINT 'Creating trigger tU_link_cache'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_cache' AND type = 'TR')
  DROP TRIGGER tU_link_cache
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
CREATE TRIGGER tU_link_cache ON link FOR UPDATE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  DECLARE @err INT, @now NVARCHAR(40), @rc INT, @verbose TINYINT
  SET @verbose = 0

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    SELECT @rc = COUNT(*) FROM inserted
    RAISERROR('%s: tU_link_cache: Starting. #rows=%d', 0, 1, @now, @rc) WITH NOWAIT
  END

  IF (UPDATE(deleted) OR UPDATE(src_cid) OR UPDATE(src_id) OR UPDATE(dst_cid) OR UPDATE(dst_id) OR UPDATE(link_type))
  BEGIN

    -- update existing old entries as deleted
    UPDATE o SET deleted = 1
    FROM ObjPathCacheLinks AS o
      JOIN deleted AS d ON  d.link_type = o.link_type
                        AND d.src_cid   = o.src_cid
                        AND d.src_id    = o.src_id
                        AND d.dst_cid   = o.dst_cid
                        AND d.dst_id    = o.dst_id
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      SELECT @rc = COUNT(*) FROM inserted
      RAISERROR('%s: tU_link_cache: Marked %d row(s) in ObjPathCacheLinks as deleted', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- treat old entries as new deleted entries
    INSERT INTO ObjPathCacheLinks (link_type, src_cid, src_id, dst_cid, dst_id, deleted)
    SELECT d.link_type, d.src_cid, d.src_id, d.dst_cid, d.dst_id, 1
    FROM ObjPathCacheLinks AS o
      RIGHT OUTER JOIN deleted AS d ON d.dst_cid   = o.dst_cid
                                   AND d.dst_id    = o.dst_id
                                   AND d.src_cid   = o.src_cid
                                   AND d.src_id    = o.src_id
                                   AND d.link_type = o.link_type
      JOIN ObjPathCacheIDs AS opc ON opc.cid = d.dst_cid
      JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = d.link_type
      JOIN isa_chain AS ic WITH (NOLOCK) ON ic.cid = d.dst_cid
    WHERE o.dst_cid IS NULL
      AND ic.base_cid = 'MNO2'
      AND ((d.dst_id <= opc.id AND d.link_type <> 'LOBL') OR (d.link_type = 'LOBL' AND d.src_id <= opc.id))
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      SELECT @rc = COUNT(*) FROM inserted
      RAISERROR('%s: tU_link_cache: Inserted %d row(s) into ObjPathCacheLinks', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- if any of the entries already exist, just mark the deleted flag to 1
    UPDATE o SET deleted = 0
    FROM ObjPathCacheLinks AS o
      JOIN inserted AS i ON o.dst_cid = i.dst_cid AND o.dst_id = i.dst_id AND o.src_cid = i.src_cid AND o.src_id = i.src_id AND i.link_type = o.link_type
    WHERE i.deleted = 0
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      SELECT @rc = COUNT(*) FROM inserted
      RAISERROR('%s: tU_link_cache: Cleared deleted flag on %d row(s) in ObjPathCacheLinks', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- add the new entries
    INSERT INTO ObjPathCacheLinks (link_type, src_cid, src_id, dst_cid, dst_id, deleted)
    SELECT RTRIM(i.link_type), RTRIM(i.src_cid), i.src_id, RTRIM(i.dst_cid), i.dst_id, i.deleted
    FROM ObjPathCacheLinks AS o
      RIGHT OUTER JOIN inserted AS i ON i.dst_cid   = o.dst_cid
                                    AND i.dst_id    = o.dst_id
                                    AND i.src_cid   = o.src_cid
                                    AND i.src_id    = o.src_id
                                    AND i.link_type = o.link_type
      JOIN ObjPathCacheIDs AS opc ON opc.cid = i.dst_cid
      JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = i.link_type
      JOIN isa_chain AS ic WITH (NOLOCK) ON ic.cid = i.dst_cid
    WHERE i.deleted = 0
      AND ((i.dst_id <= opc.id AND i.link_type <> 'LOBL') OR (i.link_type = 'LOBL' AND i.src_id <= opc.id))
      AND ic.base_cid = 'MNO2'
      AND (lt.flag_containment = 1 OR lt.flag_parent_child = 1 OR lt.flag_propagation = 1 OR (i.link_type = 'LOBL' AND i.src_cid = 'LOB' AND i.dst_cid = 'LOB'))
      AND lt.flag_aggregate = 0
      AND o.dst_cid IS NULL
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      SELECT @rc = COUNT(*) FROM inserted
      RAISERROR('%s: tU_link_cache: Inserted %d row(s) into ObjPathCacheLinks', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    SELECT @rc = COUNT(*) FROM inserted
    RAISERROR('%s: tU_link_cache: Finished', 0, 1, @now) WITH NOWAIT
  END
-- END OF tU_link_cache
GO

---------------------------------------------------------------------------------------------------
-- tU_link_LNKX
PRINT 'Creating trigger tU_link_LNKX'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_LNKX' AND type = 'TR')
  DROP TRIGGER tU_link_LNKX
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- Propagates updates that set link.deleted = 1 to LinkeState_V
CREATE TRIGGER tU_link_LNKX ON link FOR UPDATE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  DECLARE @err INT, @now NVARCHAR(40), @rc INT, @verbose TINYINT
  SET @verbose = 0

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    SELECT @rc = COUNT(*) FROM inserted
    RAISERROR('%s: tU_link_LNKX: Starting. #rows=%d', 0, 1, @now, @rc) WITH NOWAIT
  END

  IF UPDATE(deleted) AND EXISTS (SELECT 1 FROM inserted WHERE deleted = 1)
  BEGIN

    UPDATE v SET deleted = 1
    FROM LinkState_V AS v
      JOIN inserted AS i ON i.link_type = v._link_type
                        AND i.src_cid   = v._src_cid
                        AND i.src_id    = v._src_id
                        AND i.dst_cid   = v._dst_cid
                        AND i.dst_id    = v._dst_id
      JOIN deleted AS d ON d.link_type = i.link_type
                       AND d.src_cid   = i.src_cid
                       AND d.src_id    = i.src_id
                       AND d.dst_cid   = i.dst_cid
                       AND d.dst_id    = i.dst_id
    WHERE i.deleted = 1
      AND d.deleted = 0
      AND v.deleted = 0
  END

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    SELECT @rc = COUNT(*) FROM inserted
    RAISERROR('%s: tU_link_LNKX: Finished', 0, 1, @now) WITH NOWAIT
  END
-- END OF tU_link_LNKX
GO

---------------------------------------------------------------------------------------------------
-- tU_link_LOB_any_LOBL 

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_LOB_any_LOBL' AND type = 'TR')
BEGIN
  PRINT 'Dropping obsolete trigger tU_link_LOB_any_LOBL'
  DROP TRIGGER tU_link_LOB_any_LOBL
END
GO

---------------------------------------------------------------------------------------------------
-- tI_link_LOBL

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tI_link_LOBL' AND type = 'TR')
BEGIN
  PRINT 'Dropping obsolete trigger tI_link_LOBL'
  DROP TRIGGER tI_link_LOBL
END
GO

---------------------------------------------------------------------------------------------------
-- tI_link_Prop

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tI_link_Prop' AND type = 'TR')
  DROP TRIGGER tI_link_Prop
GO

PRINT 'Creating trigger tI_link_Prop'
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- This trigger is responsible for creating StagedCHEV records
-- for new link records. This happens when an existing object that has child events
-- is added to a new location in the hierarchy.
CREATE TRIGGER tI_link_Prop ON link FOR INSERT WITH APPEND
AS
  -- Table for child events to insert
  DECLARE @chev TABLE
  (
    ParentCID     NVARCHAR(4) NOT NULL,
    ParentID      INT         NOT NULL,
    ChildCID      NVARCHAR(4) NOT NULL,
    ChildID       INT         NOT NULL,
    EventCID      NVARCHAR(4) NOT NULL,
    AlertState    INT         NOT NULL,
    EventPriority INT         NOT NULL
  )

  DECLARE
    @irc      INT,
    @now      NVARCHAR(40),
    @rc       INT,
    @verbose  INT

  SET @rc = @@ROWCOUNT

  IF @rc = 0
    RETURN

  SET @verbose = 0
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tI_link_Prop: Starting. ROWCOUNT=%d', 0, 1, @now, @rc) WITH NOWAIT
  END  

  -- Create new StagedCHEV records that correspond with others in the ChildEvent table.
  -- First we see if the objects being linked to have child events. If they do then 
  -- we make sure that child events have not already been created for the new links,
  -- and for any that have not we insert into StagedCHEV.
  INSERT INTO @chev (ParentCID, ParentID, ChildCID, ChildID, EventCID, AlertState, EventPriority)
  SELECT DISTINCT i.src_cid, i.src_id, i.dst_cid, i.dst_id, c._EventCID, c._AlertState, c._EventPriority
  FROM inserted AS i
    JOIN Meta..link_type_table_BASE AS lt WITH (NOLOCK) ON lt.link_type = i.link_type
    JOIN ChildEvent_C AS c ON i.dst_cid = c._ChildCID AND i.dst_id = c._ChildID
  WHERE i.deleted = 0
    AND c.deleted = 0
    AND lt.flags & 0x0000000b > 0
  SET @rc = @@ROWCOUNT
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tI_link_Prop: Computed %d row(s) of potential ChildEvents', 0, 1, @now, @rc) WITH NOWAIT
  END
  IF @rc > 0
  BEGIN
    SET @irc = @rc
    DELETE c
    FROM @chev AS c
    JOIN ChildEvent_C AS e ON e._ChildCID      = c.ChildCID
                          AND e._ChildID       = c.ChildID
                          AND e._ParentCID     = c.ParentCID
                          AND e._ParentID      = c.ParentID
                          AND e._EventCID      = c.EventCID
                          AND e._AlertState    = c.AlertState
                          AND e._EventPriority = c.EventPriority
    WHERE e.deleted = 0
    SET @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tI_link_Prop: Removed %d row(s) from set of potential ChildEvents since they already exist', 0, 1, @now, @rc) WITH NOWAIT
    END
    IF @irc > @rc
    BEGIN
      INSERT INTO StagedCHEV (ParentCID, ParentID, ChildCID, ChildID, EventCID, AlertState, EventPriority)
      SELECT ParentCID, ParentID, ChildCID, ChildID, EventCID, AlertState, EventPriority
      FROM @chev
      SET @rc = @@ROWCOUNT
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: tI_link_Prop: Inserted %d row(s) into StagedCHEV', 0, 1, @now, @rc) WITH NOWAIT
      END
    END
  END

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tI_link_Prop: Finished', 0, 1, @now) WITH NOWAIT
  END  
GO

---------------------------------------------------------------------------------------------------
-- tU_link_Prop

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_Prop' AND type = 'TR')
  DROP TRIGGER tU_link_Prop
GO

PRINT 'Creating trigger tU_link_Prop'
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16

CREATE TRIGGER tU_link_Prop ON link FOR UPDATE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  DECLARE @err INT, @now NVARCHAR(40), @rc INT, @verbose TINYINT
  SET @verbose = 0

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    SELECT @rc = COUNT(*) FROM inserted
    RAISERROR('%s: tU_link_Prop: Starting. #rows=%d', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Handle marking a link as deleted
  IF UPDATE(deleted) AND NOT UPDATE(src_cid) AND NOT UPDATE(src_id)
  BEGIN
    UPDATE c SET deleted = 1
    FROM ChildEvent_C AS c
      JOIN deleted AS d ON d.dst_cid = c._ChildCID AND d.dst_id = c._ChildID AND d.src_cid = c._ParentCID AND d.src_id = c._ParentID
      JOIN inserted AS i ON i.dst_cid = d.dst_cid AND i.dst_id = d.dst_id
        AND i.src_cid = d.src_cid AND i.src_id = d.src_id AND i.link_type = d.link_type
      JOIN link_type AS lt ON lt.link_type = i.link_type
      LEFT OUTER JOIN link AS l ON l.dst_cid = d.dst_cid AND l.dst_id = d.dst_id AND l.src_cid = d.src_cid AND l.src_id = d.src_id
        AND l.link_type IN (SELECT link_type FROM link_type WHERE (flag_containment = 1 OR flag_parent_child = 1 OR flag_propagation = 1))
        AND l.deleted = 0
    WHERE (lt.flag_containment = 1 OR lt.flag_parent_child = 1 OR lt.flag_propagation = 1)
      -- look for deleted links
      AND i.deleted = 1 AND d.deleted = 0
      AND c.deleted = 0
      AND l.dst_cid IS NULL
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link_Prop: Marked %d ChildEvent(s) for deletion', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  -- Handle undeleting a link
  IF UPDATE(deleted) AND NOT UPDATE(src_cid) AND NOT UPDATE(src_id)
  BEGIN
    INSERT INTO StagedCHEV (ParentCID, ParentID, ChildCID, ChildID, EventCID, AlertState, EventPriority)
    SELECT DISTINCT i.src_cid, i.src_id, i.dst_cid, i.dst_id, c._EventCID, c._AlertState, c._EventPriority
    FROM ChildEvent_C AS c
      JOIN inserted AS i ON i.dst_cid = c._ChildCID AND i.dst_id = c._ChildID
      JOIN deleted AS d ON i.dst_cid = d.dst_cid AND i.dst_id = d.dst_id
        AND i.src_cid = d.src_cid AND i.src_id = d.src_id AND i.link_type = d.link_type
      JOIN link_type AS lt ON lt.link_type = i.link_type
      LEFT OUTER JOIN ChildEvent_C AS e ON e._ChildCID = i.dst_cid AND e._ChildID = i.dst_id
        AND e._ParentCID = i.src_cid AND e._ParentID = i.src_id AND e._EventCID = c._EventCID
        AND e._AlertState = c._AlertState AND e._EventPriority = c._EventPriority
    WHERE (lt.flag_containment = 1 OR lt.flag_parent_child = 1 OR lt.flag_propagation = 1)
      AND c.deleted = 0
      -- look for undeleted links
      AND i.deleted = 0 AND d.deleted = 1
      AND e._ChildCID IS NULL
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link_Prop: Inserted %d row(s) into StagedCHEV for undeleted links', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  -- Handle updating the src
  IF UPDATE(src_cid) OR UPDATE(src_id)
  BEGIN
    -- Negate old child events
    INSERT INTO StagedCHEV (ParentCID, ParentID, ChildCID, ChildID, EventCID, AlertState, EventPriority,Direction)
    SELECT DISTINCT d.src_cid, d.src_id, d.dst_cid, d.dst_id, c._EventCID, c._AlertState, c._EventPriority, -1  -- negate
    FROM inserted AS i
      JOIN deleted  AS d ON d.dst_cid = i.dst_cid AND d.dst_id = i.dst_id AND d.link_type = i.link_type
                        AND (d.src_cid <> i.src_cid OR d.src_id <> i.src_id)   -- different src
      JOIN ChildEvent_C AS c ON d.src_cid = c._ParentCID AND d.src_id = c._ParentID
                            AND d.dst_cid = c._ChildCID AND d.dst_id = c._ChildID
    WHERE d.deleted = 0 AND i.deleted = 0
      AND c.deleted = 0
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link_Prop: Inserted %d row(s) into StagedCHEV to negated from old src', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- Create new child events
    INSERT INTO StagedCHEV (ParentCID, ParentID, ChildCID, ChildID, EventCID, AlertState, EventPriority)
    SELECT DISTINCT i.src_cid, i.src_id, i.dst_cid, i.dst_id, c._EventCID, c._AlertState, c._EventPriority
    FROM inserted AS i
      JOIN deleted  AS d ON d.dst_cid = i.dst_cid AND d.dst_id = i.dst_id AND d.link_type = i.link_type
                        AND (d.src_cid <> i.src_cid OR d.src_id <> i.src_id)   -- different src
      JOIN link_type AS lt ON lt.link_type = i.link_type
      JOIN ChildEvent_C AS c ON i.dst_cid = c._ChildCID AND i.dst_id = c._ChildID
      LEFT OUTER JOIN ChildEvent_C AS e ON e._ChildCID = i.dst_cid AND e._ChildID = i.dst_id
        AND e._ParentCID = i.src_cid AND e._ParentID = i.src_id AND e._EventCID = c._EventCID
        AND e._AlertState = c._AlertState AND e._EventPriority = c._EventPriority
        AND e.deleted = 0
    WHERE d.deleted = 0 AND i.deleted = 0
      AND (lt.flag_containment = 1 OR lt.flag_parent_child = 1 OR lt.flag_propagation = 1)
      AND c.deleted = 0
      AND e._ChildCID IS NULL
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link_Prop: Inserted %d row(s) into StagedCHEV for new src', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tU_link_Prop: Finished', 0, 1, @now) WITH NOWAIT
  END
GO

---------------------------------------------------------------------------------------------------
-- tD_link_Prop

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tD_link_Prop' AND type = 'TR')
  DROP TRIGGER tD_link_Prop
GO

PRINT 'Creating trigger tD_link_Prop'
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16

CREATE TRIGGER tD_link_Prop ON link FOR DELETE WITH APPEND
AS

  DECLARE @rc INT
  SET @rc = @@ROWCOUNT

  IF @rc = 0
    RETURN

  -- Negate child events for links that are not deleted
  UPDATE c
  SET c.deleted = 1
  FROM ChildEvent_C AS c
    JOIN deleted AS d ON d.dst_cid = c._ChildCID AND d.dst_id = c._ChildID AND d.src_cid = c._ParentCID AND d.src_id = c._ParentID
    JOIN link_type AS lt ON lt.link_type = d.link_type
    LEFT OUTER JOIN link AS l ON l.dst_cid = d.dst_cid AND l.dst_id = d.dst_id AND l.src_cid = d.src_cid AND l.src_id = d.src_id
      AND l.link_type IN (SELECT link_type FROM link_type WHERE (flag_containment = 1 OR flag_parent_child = 1 OR flag_propagation = 1))
      AND l.deleted = 0
  WHERE (lt.flag_containment = 1 OR lt.flag_parent_child = 1 OR lt.flag_propagation = 1)
    AND d.deleted = 0
    AND c.deleted = 0
    AND l.dst_cid IS NULL
GO

---------------------------------------------------------------------------------------------------
-- tI_link_Sysplex

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tI_link_Sysplex' AND type = 'TR')
  DROP TRIGGER tI_link_Sysplex
GO

PRINT 'Creating trigger tI_link_Sysplex'
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- This trigger rebuilds the OS_SYSPLEX table
CREATE TRIGGER tI_link_Sysplex ON link FOR INSERT WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  IF EXISTS (SELECT 1 FROM inserted WHERE link_type = 'LOBL' AND dst_cid = 'OS')
  BEGIN
    -- An OS is being placed into a Business System, so we'll rebuild the OS_SYSPLEX table
    -- NOTE: It is possible that the target Business System is not related to the SYSPLEX LOB structure.
    --       However it is quicker to just rebuild the OS_SYSPLEX than first try to determine if the Insert is
    --       into the SYSPLEX LOB structure
    EXEC asisp_populateOS_SYSPLEX
  END

GO

---------------------------------------------------------------------------------------------------
-- tU_link_Sysplex

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_Sysplex' AND type = 'TR')
  DROP TRIGGER tU_link_Sysplex
GO

PRINT 'Creating trigger tU_link_Sysplex'
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- This trigger rebuilds the OS_SYSPLEX table
CREATE TRIGGER tU_link_Sysplex ON link FOR UPDATE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  -- Handle marking a link as deleted
  IF UPDATE(deleted) AND EXISTS (SELECT 1 FROM inserted WHERE dst_cid = 'OS')
  BEGIN
    -- An OS is being marked as deleted
    -- NOTE: It is possible that this delete is not related to the SYSPLEX LOB structure or physical OS.
    --       However it is quicker to just rebuild the OS_SYSPLEX than first try to determine if it is related.
    EXEC asisp_populateOS_SYSPLEX
  END

GO

---------------------------------------------------------------------------------------------------
-- tD_link_Sysplex

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tD_link_Sysplex' AND type = 'TR')
  DROP TRIGGER tD_link_Sysplex
GO

PRINT 'Creating trigger tD_link_Sysplex'
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
-- This trigger rebuilds the OS_SYSPLEX table
CREATE TRIGGER tD_link_Sysplex ON link FOR DELETE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  IF EXISTS (SELECT 1 FROM deleted WHERE dst_cid = 'OS')
  BEGIN
    -- An OS is being deleted
    -- NOTE: It is possible that this delete is not related to the SYSPLEX LOB structure or physical OS.
    --       However it is quicker to just rebuild the OS_SYSPLEX than first try to determine if it is related.
    EXEC asisp_populateOS_SYSPLEX
  END
GO

---------------------------------------------------------------------------------------------------
-- tI_link_PBT

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tI_link_PBT' AND type = 'TR')
BEGIN
  PRINT 'Dropping obsolete trigger tI_link_PBT'
  DROP TRIGGER tI_link_PBT
END
GO

---------------------------------------------------------------------------------------------------
-- tU_link_PBT
PRINT 'Creating trigger tU_link_PBT'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_link_PBT' AND type = 'TR')
  DROP TRIGGER tU_link_PBT
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 2003,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive\db\sql\tables\link.sql
-- SID: 1.15.2.36
-- Date Modified: 04/12/13 15:25:16
CREATE TRIGGER tU_link_PBT ON link FOR UPDATE WITH APPEND
AS
  IF @@ROWCOUNT = 0
    RETURN

  DECLARE @err INT, @now NVARCHAR(40), @rc INT, @verbose TINYINT
  SET @verbose = 0

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    SELECT @rc = COUNT(*) FROM inserted
    RAISERROR('%s: tU_link_PBT: Starting. #rows=%d', 0, 1, @now, @rc) WITH NOWAIT
  END

  IF EXISTS (SELECT 1 FROM inserted WHERE link_type = 'LOBC' AND src_cid = 'LOB')
  BEGIN
    -- The src_cid/src_id columns of the updated rows are joined to PBTUsedBy obj_cid/obj_id.
    -- PBTUsedBy is joined to PercentageBasedThreshold to filter-out PBTs that are not enabled.
    INSERT INTO StagedPBT ( id )
      SELECT DISTINCT u.id
      FROM inserted AS i
      JOIN deleted AS d ON d.link_type = i.link_type 
        AND d.src_cid = i.src_cid
        AND d.src_id  = i.src_id
        AND d.dst_cid = i.dst_cid
        AND d.dst_id  = i.dst_id
      JOIN PBTUsedBy AS u ON u.obj_cid = i.src_cid AND u.obj_id = i.src_id
      JOIN PercentageBasedThreshold AS p ON p.id = u.pbt_id
      WHERE i.link_type = 'LOBC'
        AND i.src_cid   = 'LOB'
        AND i.dst_cid   = 'LOB'
        AND i.deleted  <> d.deleted
        AND u.enabled   = 1
        AND p.enabled   = 1
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link_PBT: Inserted %d row(s) into StagedPBT', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- The src_cid/src_id columns of the inserted rows are joined to lob_link to get shorcut
    -- LOBs that need PBT evaluations
    INSERT INTO StagedPBT ( id )
      SELECT DISTINCT u.id
      FROM inserted AS i
      JOIN deleted AS d ON d.link_type = i.link_type 
        AND d.src_cid = i.src_cid
        AND d.src_id  = i.src_id
        AND d.dst_cid = i.dst_cid
        AND d.dst_id  = i.dst_id
      JOIN lob_link  AS l WITH (NOLOCK) ON l.phy_cid = i.src_cid AND l.phy_id = i.src_id
      JOIN PBTUsedBy AS u ON u.obj_id = l.dst_id
      JOIN PercentageBasedThreshold AS p ON p.id = u.pbt_id
      WHERE i.link_type = 'LOBC'
        AND i.src_cid   = 'LOB'
        AND i.dst_cid   = 'LOB'
        AND i.deleted  <> d.deleted
        AND u.enabled   = 1
        AND p.enabled   = 1
        AND u.obj_cid   = 'LOB'
        AND l.src_id    = 0
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: tU_link_PBT: Inserted %d row(s) into StagedPBT for BS Shortcuts', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: tU_link_PBT: Finished', 0, 1, @now) WITH NOWAIT
  END
GO
